#!/usr/bin/env python3
"""
http://blog.nodraak.fr/2020/10/symbolic-maths-in-python/
"""

from matplotlib import pyplot as plt
import numpy as np
import sympy as sp


MATPLOTLIB_FIGSIZE = (2*6.4, 2*4.8)


def deg2rad(d):
    return d*sp.pi/180


def rad2deg(d):
    return d*180/sp.pi


#
# 1.1 Symbols
#

def example_1_1_symbols():
    x = sp.symbols('x')

    expr = sp.cos(x)**2 + sp.sin(x)**2

    print(expr)                             # Shows: sin(x)**2 + cos(x)**2
    print(sp.simplify(expr))                # Shows: 1


#
# 1.2a Mathematical operations a
#

def example_1_2a_math_expr_a():
    x, a, b = sp.symbols('x a b')

    # definite integral of x**2 between 0 and 1
    expr = x**2
    integrated = sp.integrate(expr, (x, 0, 1))
    print(integrated)                       # Shows: 1/3

    # indefinite integral of ax+b
    expr = a*x + b
    integrated = sp.integrate(expr, x)
    print(integrated)                       # Shows: a*x**2/2 + b*x


#
# 1.2b Mathematical operations b
#

def example_1_2b_math_expr_b():
    x = sp.symbols('x')

    expr = x**2
    integrated = sp.integrate(expr, x)
    print(integrated)                       # Shows: x**3/3

    value_at_1 = integrated.subs(x, 1)
    print(value_at_1)                       # Shows: 1/3


#
# 2 Cannonball shot
#

def example_2a_cannonball_shot_eq():
    t = sp.symbols('t', real=True, positive=True)
    g = sp.symbols('g', real=True, positive=True)
    v0, angle = sp.symbols('v0 angle', real=True, positive=True)

    G = 9.81

    acc_x = 0
    acc_y = -g

    vel_x = sp.integrate(acc_x, t) + v0*sp.cos(angle)
    vel_y = sp.integrate(acc_y, t) + v0*sp.sin(angle)

    pos_x = sp.integrate(vel_x, t)
    pos_y = sp.integrate(vel_y, t)

    print('acc_x =', acc_x)                 # Shows: acc_x = 0
    print('acc_y =', acc_y)                 # Shows: acc_y = -g
    print('vel_x =', vel_x)                 # Shows: vel_x = v0*cos(angle)
    print('vel_y =', vel_y)                 # Shows: vel_y = -g*t + v0*sin(angle)
    print('pos_x =', pos_x)                 # Shows: pos_x = t*v0*cos(angle)
    print('pos_y =', pos_y)                 # Shows: pos_y = -g*t**2/2 + t*v0*sin(angle)

    return {
        't': t,
        'g': g,
        'v0': v0,
        'G': G,
        'angle': angle,
        'pos_x': pos_x,
        'pos_y': pos_y,
    }


def example_2b_calc_traj(common, V0, ANGLE, TMAX):
    t, g, v0, angle = common['t'], common['g'], common['v0'], common['angle']
    G = common['G']
    pos_x, pos_y = common['pos_x'], common['pos_y']

    # Simplify a bit the equation now

    pos_x_t = pos_x.subs(v0, V0).subs(angle, ANGLE)
    pos_y_t = pos_y.subs(g, G).subs(v0, V0).subs(angle, ANGLE)

    print(pos_x_t)                          # Shows: 37.1579645676346*t
    print(pos_y_t)                          # Shows: -4.905*t**2 + 33.4557270013732*t

    # Compute the trajectory until the cannonball hits the ground, with a 0.1 sec increment

    pxs, pys = [], []

    for T in np.linspace(0, TMAX, TMAX*10+1):
        px = float(pos_x_t.subs(t, T))
        py = float(pos_y_t.subs(t, T))

        # we have hit the ground? Then break
        if py < 0:
            break

        pxs.append(px)
        pys.append(py)

    return pxs, pys


def example_2c_cannonball_shot_plot(common):
    V0 = 50
    TMAX = 200

    plt.figure(figsize=MATPLOTLIB_FIGSIZE)

    plt.gca().axhline(0, color='gray')
    plt.vlines(0, 0, 100, color='gray')

    legend_labels = []
    legend_handles = []

    for ANGLE_DEG in range(20, 50+1, 10):
        ANGLE = deg2rad(ANGLE_DEG)

        pxs, pys = example_2b_calc_traj(common, V0, ANGLE, TMAX)

        (h_v0, ) = plt.plot([0, V0*sp.cos(ANGLE)], [0, V0*sp.sin(ANGLE)], '--')
        (h_dots, ) = plt.plot(pxs, pys, color=h_v0.get_color())

        legend_handles.append(h_dots)
        legend_labels.append('%d°' % ANGLE_DEG)

    plt.legend(legend_handles, legend_labels)

    plt.axis('equal')
    plt.tight_layout()
    plt.savefig('example_2c_cannonball_shot_plot.png')
    plt.show()


#
# 3 Attack
#

def _example_3_plot_castle():

    # First, some constants and helpers

    CASTLE_COLOR = 'gray'
    CASTLE_POS_X = 200
    TOWER_BASE_HEIGHT = 30
    TOWER_BASE_WIDTH = 10
    TOWER_OVERHANG = 3
    TOWER_CRENEL_COUNT = 3

    GRASS_COLOR = 'green'


    def plt_line(x1, y1, x2, y2, **kwargs):
        return plt.plot([x1, x2], [y1, y2], **kwargs)


    def plot_tower(x):
        # vert
        plt_line(x, 0, x, TOWER_BASE_HEIGHT, color=CASTLE_COLOR)
        plt_line(x+TOWER_BASE_WIDTH, 0, x+TOWER_BASE_WIDTH, TOWER_BASE_HEIGHT, color=CASTLE_COLOR)
        # top
        plt_line(x-TOWER_OVERHANG, TOWER_BASE_HEIGHT, x+TOWER_BASE_WIDTH+TOWER_OVERHANG, TOWER_BASE_HEIGHT, color=CASTLE_COLOR)
        # crenels
        tower_crenel_size = (TOWER_BASE_WIDTH+2*TOWER_OVERHANG)/(TOWER_CRENEL_COUNT+0.5)
        base_crenel_y = TOWER_BASE_HEIGHT+3+tower_crenel_size
        plt_line(x-TOWER_OVERHANG, TOWER_BASE_HEIGHT, x-TOWER_OVERHANG, base_crenel_y, color=CASTLE_COLOR)
        plt_line(x+TOWER_BASE_WIDTH+TOWER_OVERHANG, TOWER_BASE_HEIGHT, x+TOWER_BASE_WIDTH+TOWER_OVERHANG, base_crenel_y, color=CASTLE_COLOR)

        for i in range(int(TOWER_CRENEL_COUNT)):
            base_crenel_x = x-TOWER_OVERHANG + i*tower_crenel_size
            plt.plot([
                    base_crenel_x,
                    base_crenel_x+0.5*tower_crenel_size,
                    base_crenel_x+0.5*tower_crenel_size,
                    base_crenel_x+tower_crenel_size,
                    base_crenel_x+tower_crenel_size,
                ],
                [
                    base_crenel_y,
                    base_crenel_y,
                    base_crenel_y-0.5*tower_crenel_size,
                    base_crenel_y-0.5*tower_crenel_size,
                    base_crenel_y,
                ],
                color=CASTLE_COLOR,
            )
        base_crenel_x = x-TOWER_OVERHANG + (i+1)*tower_crenel_size
        base_crenel_y = TOWER_BASE_HEIGHT+3+tower_crenel_size
        plt.plot([
                base_crenel_x,
                base_crenel_x+0.5*tower_crenel_size,
            ],
            [
                base_crenel_y,
                base_crenel_y,
            ],
            color=CASTLE_COLOR,
        )

    # Then, the actual function

    # base
    plt_line(CASTLE_POS_X, 0, CASTLE_POS_X+2*TOWER_BASE_WIDTH+30, 0, color=CASTLE_COLOR)
    # towers
    plot_tower(CASTLE_POS_X)
    plot_tower(CASTLE_POS_X+TOWER_BASE_WIDTH+30)
    # center top
    plt_line(CASTLE_POS_X+TOWER_BASE_WIDTH, 20, CASTLE_POS_X+TOWER_BASE_WIDTH+30, 20, color=CASTLE_COLOR)
    # grass
    plt_line(-20, 0, 270, 0, color=GRASS_COLOR)


def example_3a_attack_castle():
    TARGET_X = 245
    TARGET_Y = 36.5

    plt.figure(figsize=MATPLOTLIB_FIGSIZE)

    plt.plot(0, 0, 'x', color='green')
    plt.plot(TARGET_X, TARGET_Y, 'x', color='red')

    _example_3_plot_castle()

    plt.axis('equal')
    plt.tight_layout()
    plt.savefig('example_3a_attack_castle.png')
    plt.show()


def example_3b_attack_v0_60(common):
    t, g, v0, angle = common['t'], common['g'], common['v0'], common['angle']
    G = common['G']
    pos_x, pos_y = common['pos_x'], common['pos_y']

    V0 = 60
    TARGET_X = 245
    TARGET_Y = 36.5

    # Let's do so substitutions/simplifications before the heavy calculation
    pos_x_t = pos_x.subs(v0, V0)
    pos_y_t = pos_y.subs(g, G).subs(v0, V0)

    # The meat of the solver is here. The two arrays passed as arguments are:
    # 1. Equations to solve ("sp.Eq(pos_x_t, TARGET_X)" means "pos_x_t == TARGET_X")
    # 2. Variables to solve for
    sols = sp.solve([
        sp.Eq(pos_x_t, TARGET_X),
        sp.Eq(pos_y_t, TARGET_Y),
    ], [angle, t])

    # Let's get all the valid solutions
    # When declaring SymPy symbols, we specified that angle and t were positive, so
    # all solutions should be valid (no negative times or velocity).
    sols_v0_60 = []  # angle_rad, t
    for (angle_sol, t_sol) in sols:
        sols_v0_60.append((float(angle_sol), float(t_sol)))

    # Print the solutions
    print('v0: angle_deg t')
    for (ANGLE_RAD, T) in sols_v0_60:
        print('%d: %d %4.1f' % (V0, rad2deg(ANGLE_RAD), T))

    # The above code prints:
    # v0: angle_deg t
    # 60: 31  4.8
    # 60: 67 10.6

    return sols_v0_60


def example_3c_attack_v0_min(common):
    t, g, v0, angle = common['t'], common['g'], common['v0'], common['angle']
    G = common['G']
    pos_x, pos_y = common['pos_x'], common['pos_y']

    TARGET_X = 245
    TARGET_Y = 36.5

    pos_y = pos_y.subs(g, G)

    # Similar as before, solve for v0
    sols = sp.solve([
        sp.Eq(pos_x, TARGET_X),
        sp.Eq(pos_y, TARGET_Y),
    ], [v0, angle])
    assert len(sols) == 1
    v0_sol, angle_sol = sols[0]

    # Once we have v0, solve for t when its derivative over t is equal to 0
    sols = sp.solve([sp.diff(v0_sol, t)], [t])
    assert len(sols) == 1
    (t_sol, ) = sols[0]

    # Finally, reduce all the variables
    v0_min = float(v0_sol.subs(t, t_sol))
    angle_at_v0min = float(angle_sol.subs(t, t_sol))
    t_at_v0min = float(t_sol)

    print('v0_min = %.1f' % v0_min)                     # Shows: v0_min = 52.8
    print('angle_deg = %.1f' % rad2deg(angle_at_v0min)) # Shows: angle_deg = 49.2
    print('t = %.1f' % t_at_v0min)                      # Shows: t = 7.1

    return (v0_min, angle_at_v0min, t_at_v0min)


def example_3d_attack_shots(common, sols_v0_60, sols_min):
    t, g, v0, angle = common['t'], common['g'], common['v0'], common['angle']
    G = common['G']
    pos_x, pos_y = common['pos_x'], common['pos_y']

    # img setup

    plt.figure(figsize=MATPLOTLIB_FIGSIZE)

    legends = []  # (sorting_key, handle, label)

    # castle

    _example_3_plot_castle()

    # shot at v0=60

    V0 = 60
    for (ANGLE_RAD, T) in sols_v0_60:
        pxs, pys = example_2b_calc_traj(common, V0, ANGLE_RAD, T)

        (h_v0, ) = plt.plot([0, V0*sp.cos(ANGLE_RAD)], [0, V0*sp.sin(ANGLE_RAD)], '--')
        (h_traj, ) = plt.plot(pxs, pys, color=h_v0.get_color())

        legends.append((ANGLE_RAD, h_traj, '%d° => %d m/s / %.1f sec' % (rad2deg(ANGLE_RAD), V0, T)))

    # shot at v0 min

    # TODO: from example_3c_attack_v0_min
    V0, ANGLE_RAD, T = sols_min

    pxs, pys = example_2b_calc_traj(common, V0, ANGLE_RAD, T)

    (h_v0, ) = plt.plot([0, V0*sp.cos(ANGLE_RAD)], [0, V0*sp.sin(ANGLE_RAD)], '--')
    (h_traj, ) = plt.plot(pxs, pys, color=h_v0.get_color())

    legends.append((ANGLE_RAD, h_traj, '%d° => %d m/s / %.1f sec' % (rad2deg(ANGLE_RAD), V0, T)))

    # legend and stuff

    legends = sorted(legends, key=lambda tup: tup[0])
    legends_handles = [tup[1] for tup in legends]
    legends_labels = [tup[2] for tup in legends]
    plt.legend(legends_handles, legends_labels)

    plt.axis('equal')
    plt.tight_layout()
    plt.savefig('example_3b_attack_shots.png')
    plt.show()


#
# main
#

def main():

    print('=== example_1_1_symbols() ===')
    example_1_1_symbols()


    print('=== example_1_2a_math_expr_a() ===')
    example_1_2a_math_expr_a()

    print('=== example_1_2b_math_expr_b() ===')
    example_1_2b_math_expr_b()


    print('=== example_2a_cannonball_shot_eq() ===')
    common = example_2a_cannonball_shot_eq()

    print('=== example_2b_calc_traj() ===')
    example_2b_calc_traj(common, 50, deg2rad(42), 200)

    print('=== example_2c_cannonball_shot_plot() ===')
    example_2c_cannonball_shot_plot(common)


    print('=== example_3a_attack_castle() ===')
    example_3a_attack_castle()

    print('=== example_3b_attack_v0_60() ===')
    sols_v0_60 = example_3b_attack_v0_60(common)

    print('=== example_3c_attack_v0_min() ===')
    sols_min = example_3c_attack_v0_min(common)

    print('=== example_3d_attack_shots() ===')
    example_3d_attack_shots(common, sols_v0_60, sols_min)


main()
